/* par_sym_stack.c */
/*
   Copyright (C) 2006 John W. Jordan
   All Rights Reserved

  This file is part of Dante.
  Dante is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the license, or
  ( at your option ) any later version.
  Dante is distributed in the hope that it will be usefull,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A OARTICULAR PURPOSE. See the
  GNU General Public License for more detaile.
  You should have received a copy of the GNU General Public License
  along with Dante; if not write to the Free Software
  Foundation. Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
USA
*/

/* Uses a stack as a symbol table. */
#include <stdio.h>
#include <string.h>
#include "par_stack.h"
#include "par_sym.h"

/* ------------------------------------------------ */
static int sym_count_res_words (STACK * stack);

/* --------------------------------------------------------- */
/* Add sorted sym to stack. */
/*
   Adding to the stack in order requires
   moving data since the stack has the form
   of an array.
*/
SYM *
sym_add (STACK * stack, char *name)
{
  SYM *sym;
  SYM *e;
  SYM *x;
  int k;
  int j;

  if (sym = sym_find_symbol (stack, name))
    return sym;

  /* Default to reserved symbol. */
  sym = (SYM *) xmalloc (sizeof (SYM));
  sym->type = -1;
  sym->name = strdup (name);
  sym->reserved = 0;
  sym->flex_def = 0;

  /* Push onto stack to asssure there is room. */
  /* If stack was empty, just return. */

  Push (stack, sym);

  /* Put in sorted position. */
  x = sym;
  for (k = 0; k < stack->no; k++)
    {
      e = stack->item[k];
      if (strcmp (x->name, e->name) < 0)
	break;
    }
  /* Move others down in array. */
  for (j = stack->no - 1; j > k; j--)
    stack->item[j] = stack->item[j - 1];

  /* Insert */
  stack->item[k] = x;

  return sym;
}

/* ------------------------------------------------ */
char *
sym_name (SYM * sym)
{
  return sym->name;
}

/* ------------------------------------------------ */
int
sym_type_code (SYM * sym)
{
  return sym->type;
}

/* ------------------------------------------------ */
void
sym_set_type (SYM * sym, int type)
{
  sym->type = type;
}

/* ------------------------------------------------ */
char *
sym_set_res (SYM * sym, int res)
{
  sym->reserved = res;
}

/* ------------------------------------------------ */
void
sym_add_flexdef (STACK * stack, char *name, char *flex_def)
{
  SYM *sym;
  if ((sym = sym_find_symbol (stack, name)) != 0)
    sym->flex_def = strdup (flex_def);
  else
    {
      fprintf (stderr, "sym_add_flexdef:\n");
      fprintf (stderr, "ERROR:SYMTBL:Name [%s] not in symtbl.\n", name);
      fprintf (stderr, "ERROR:Did not add flex def [%s].\n", flex_def);
    }
}

/* ------------------------------------------------ */
SYM *
sym_find_symbol (STACK * stack, char *name)
{
  SYM *sym;
  int size;
  int i = 0;
  int j;
  int k;

  size = stack->no;

  if (size == 0)
    return 0;

  j = size - 1;

  while (i <= j)
    {
      k = (i + j) >> 1;
      sym = (SYM *) stack->item[k];
      switch (cmpstr (name, sym->name))
	{
	case -1:
	  j = k - 1;
	  break;
	case 0:
	  return sym;
	  break;
	case 1:
	  i = k + 1;
	  break;
	}
    }
  return 0;
}

/* ------------------------------------------------ */
SYM *
sym_set_type_by_name (STACK * stack, char *name, int type)
{
  SYM *sym;
  int k;

  for (k = 0; k < stack->no; k++)
    {
      sym = (SYM *) stack->item[k];
      if (strcmp (sym->name, name) == 0)
	{
	  sym->type = type;
	  return sym;
	}
    }
  return 0;
}

/* --------------------------------------------------------- */
void
sym_pr_symtbl (STACK * stack)
{
  SYM *sym;
  int k;
  if (stack == 0)
    return;
  printf ("RESERVED WORDS:\n");
  for (k = 0; k < stack->no; k++)
    {
      sym = (SYM *) stack->item[k];
      if (sym->reserved)
	{
	  printf ("reswd name %s type_code %d\n", sym->name, sym->type);
	}
    }
  printf ("TYPES - TO BE DEFINED BY LEXER:\n");
  for (k = 0; k < stack->no; k++)
    {
      sym = (SYM *) stack->item[k];
      if (!sym->reserved)
	{
	  if (sym->flex_def)
	    printf ("%s flex/lex def: %s\n", sym->name, sym->flex_def);
	  printf ("type name %s type_code %d", sym->name, sym->type);
	  if (sym->flex_def)
	    printf ("%s flex_def %s\n", sym->flex_def);
	  else
	    printf ("\n");
	}
    }
}

/* --------------------------------------------------------- */
int
sym_set_res_table_codes (STACK * stack, int code)
{
  SYM *sym;
  int k;

  if (stack == 0)
    return;

  for (k = 0; k < stack->no; k++)
    {
      sym = (SYM *) stack->item[k];
      if (sym->type == (-1))
	sym->type = code++;
    }
  return code;
}

/* --------------------------------------------------------- */
int
sym_set_opr_table_codes (STACK * stack, int code)
{
  SYM *sym;
  int k;

  if (stack == 0)
    return;

  for (k = 0; k < stack->no; k++)
    {
      sym = (SYM *) stack->item[k];
      if (sym->type == (-1))
	{
	  if (strlen (sym->name) == 1)
	    sym->type = sym->name[0];
	  else
	    sym->type = code++;
	}
    }
  return code;
}

/* --------------------------------------------------------- */
void
sym_pr_stack (STACK * stack)
{
  SYM *sym;
  int k;

  if (stack == 0)
    return;

  for (k = 0; k < stack->no; k++)
    {
      sym = (SYM *) stack->item[k];
      printf ("%s\n", sym->name);
    }
  printf ("\n");
}

/* --------------------------------------------------------- */
int
sym_isres (STACK * stack, char *name)
{
  SYM *sym;

  sym = sym_find_symbol (stack, name);
  return sym->reserved;
}

/* --------------------------------------------------------- */
void
sym_rwtbl (STACK * stack, FILE * fd)
{
  SYM *sym;
  int k;
  int rwtbl_size = sym_count_res_words (stack);

  fprintf (fd, "typedef struct \n");
  fprintf (fd, "{ \n");
  fprintf (fd, "    char   *name;\n");
  fprintf (fd, "    int    type;\n");
  fprintf (fd, "}RWTBL; \n\n");
  fprintf (fd, "#define RWTBL_SIZE %d\n", rwtbl_size);
  fprintf (fd, "static const RWTBL rwtbl[RWTBL_SIZE] = \n");
  fprintf (fd, "{\n");

  for (k = 0; k < stack->no; k++)
    {
      sym = (SYM *) stack->item[k];
      if (sym->reserved)
	fprintf (fd, "   {\"%s\",\t%d },\n", sym->name, sym->type);
    }
  fprintf (fd, "};\n");
}

/* --------------------------------------------------------- */
void
sym_optbl (STACK * stack, FILE * fd)
{
  SYM *sym;
  int k;
  int optbl_size = sym_count_res_words (stack);

  fprintf (fd, "typedef struct \n");
  fprintf (fd, "{ \n");
  fprintf (fd, "    char   *name;\n");
  fprintf (fd, "    int    type;\n");
  fprintf (fd, "}OPTBL; \n\n");
  fprintf (fd, "#define OPTBL_SIZE %d\n", optbl_size);
  fprintf (fd, "static const OPTBL optbl[OPTBL_SIZE] = \n");
  fprintf (fd, "{\n");

  for (k = 0; k < stack->no; k++)
    {
      sym = (SYM *) stack->item[k];
      fprintf (fd, "   {\"%s\",\t%d },\n", sym->name, sym->type);
    }
  fprintf (fd, "};\n");
}

/* --------------------------------------------------------- */
static int
sym_count_res_words (STACK * stack)
{
  SYM *sym;
  int count = 0;
  int k;

  for (k = 0; k < stack->no; k++)
    {
      sym = (SYM *) stack->item[k];
      if (sym->reserved)
	count++;
    }
  return count;
}
